@@ -1,16 +1,19 @@
 import array
+import errno
 import io
 import socket
 import sys
 from struct import Struct
 from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
 
-from ..constants import MESSAGE_FLAG_MAP, MESSAGE_TYPE_MAP
+from ..constants import MESSAGE_FLAG_MAP, MESSAGE_TYPE_MAP, MessageFlag
 from ..errors import InvalidMessageError
 from ..message import Message
 from ..signature import SignatureType, Variant, get_signature_tree
 from .constants import BIG_ENDIAN, LITTLE_ENDIAN, PROTOCOL_VERSION
 
+MESSAGE_FLAG_INTENUM = MessageFlag
+
 MAX_UNIX_FDS = 16
 MAX_UNIX_FDS_SIZE = array.array("i").itemsize
 UNIX_FDS_CMSG_LENGTH = socket.CMSG_LEN(MAX_UNIX_FDS_SIZE)
@@ -102,6 +105,10 @@ ARRAY = array.array
 SOL_SOCKET = socket.SOL_SOCKET
 SCM_RIGHTS = socket.SCM_RIGHTS
 
+EAGAIN = errno.EAGAIN
+EWOULDBLOCK = errno.EWOULDBLOCK
+
+
 HEADER_MESSAGE_ARG_NAME = {
     1: "path",
     2: "interface",
@@ -115,11 +122,14 @@ HEADER_MESSAGE_ARG_NAME = {
 }
 
 _SignatureType = SignatureType
+_int = int
 
 READER_TYPE = Callable[["Unmarshaller", SignatureType], Any]
 
 MARSHALL_STREAM_END_ERROR = BlockingIOError
 
+DEFAULT_BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE
+
 
 def unpack_parser_factory(unpack_from: Callable, size: int) -> READER_TYPE:
     """Build a parser that unpacks the bytes using the given unpack_from function."""
@@ -150,6 +160,7 @@ try:
 except ImportError:
     from ._cython_compat import FAKE_CYTHON as cython
 
+
 #
 # Alignment padding is handled with the following formula below
 #
@@ -168,6 +179,12 @@ except ImportError:
 #
 #
 class Unmarshaller:
+    """Unmarshall messages from a stream.
+
+    When calling with sock and _negotiate_unix_fd False, the unmashaller must
+    be called continuously for each new message as it will buffer the data
+    until a complete message is available.
+    """
 
     __slots__ = (
         "_unix_fds",
@@ -188,9 +205,18 @@ class Unmarshaller:
         "_uint16_unpack",
         "_is_native",
         "_stream_reader",
+        "_sock_reader",
+        "_negotiate_unix_fd",
+        "_read_complete",
+        "_endian",
     )
 
-    def __init__(self, stream: io.BufferedRWPair, sock: Optional[socket.socket] = None):
+    def __init__(
+        self,
+        stream: Optional[io.BufferedRWPair] = None,
+        sock: Optional[socket.socket] = None,
+        negotiate_unix_fd: bool = True,
+    ) -> None:
         self._unix_fds: List[int] = []
         self._buf = bytearray()  # Actual buffer
         self._stream = stream
@@ -209,25 +235,29 @@ class Unmarshaller:
         self._int16_unpack: Optional[Callable] = None
         self._uint16_unpack: Optional[Callable] = None
         self._stream_reader: Optional[Callable] = None
-        if self._sock is None:
+        self._negotiate_unix_fd = negotiate_unix_fd
+        self._read_complete = False
+        if stream:
             if isinstance(stream, io.BufferedRWPair) and hasattr(stream, "reader"):
                 self._stream_reader = stream.reader.read  # type: ignore[attr-defined]
             self._stream_reader = stream.read
+        elif self._negotiate_unix_fd:
+            self._sock_reader = self._sock.recvmsg
+        else:
+            self._sock_reader = self._sock.recv
+        self._endian = 0
 
-    def reset(self) -> None:
-        """Reset the unmarshaller to its initial state.
-
-        Call this before processing a new message.
-        """
-        self._reset()
-
-    def _reset(self) -> None:
+    def _next_message(self) -> None:
         """Reset the unmarshaller to its initial state.
 
         Call this before processing a new message.
         """
         self._unix_fds = []
-        self._buf.clear()
+        to_clear = HEADER_SIGNATURE_SIZE + self._msg_len
+        if len(self._buf) == to_clear:
+            self._buf = bytearray()
+        else:
+            del self._buf[:to_clear]
         self._message = None
         self._pos = 0
         self._body_len = 0
@@ -237,6 +267,7 @@ class Unmarshaller:
         self._flag = 0
         self._msg_len = 0
         self._is_native = 0
+        self._read_complete = False
         # No need to reset the unpack functions, they are set in _read_header
         # every time a new message is processed.
 
@@ -245,22 +276,75 @@ class Unmarshaller:
         """Return the message that has been unmarshalled."""
         return self._message
 
-    def _read_sock(self, length: int) -> bytes:
+    def _has_another_message_in_buffer(self) -> bool:
+        """Check if there is another message in the buffer."""
+        return len(self._buf) > HEADER_SIGNATURE_SIZE + self._msg_len
+
+    def _read_sock_with_fds(self, pos: _int, missing_bytes: _int) -> None:
         """reads from the socket, storing any fds sent and handling errors
-        from the read itself"""
+        from the read itself.
+
+        This function is greedy and will read as much data as possible
+        from the underlying socket.
+        """
         # This will raise BlockingIOError if there is no data to read
         # which we store in the MARSHALL_STREAM_END_ERROR object
-        msg, ancdata, _flags, _addr = self._sock.recvmsg(length, UNIX_FDS_CMSG_LENGTH)  # type: ignore[union-attr]
-        for level, type_, data in ancdata:
-            if not (level == SOL_SOCKET and type_ == SCM_RIGHTS):
-                continue
-            self._unix_fds.extend(
-                ARRAY("i", data[: len(data) - (len(data) % MAX_UNIX_FDS_SIZE)])
-            )
+        try:
+            recv = self._sock_reader(missing_bytes, UNIX_FDS_CMSG_LENGTH)  # type: ignore[union-attr]
+        except OSError as e:
+            errno = e.errno
+            if errno == EAGAIN or errno == EWOULDBLOCK:
+                raise MARSHALL_STREAM_END_ERROR
+            raise
+        msg = recv[0]
+        ancdata = recv[1]
+        if ancdata:
+            for level, type_, data in ancdata:
+                if not (level == SOL_SOCKET and type_ == SCM_RIGHTS):
+                    continue
+                self._unix_fds.extend(
+                    ARRAY("i", data[: len(data) - (len(data) % MAX_UNIX_FDS_SIZE)])
+                )
+        if msg == b"":
+            raise EOFError()
+        self._buf += msg
+        if len(self._buf) < pos:
+            raise MARSHALL_STREAM_END_ERROR
 
-        return msg
+    def _read_sock_without_fds(self, pos: _int) -> None:
+        """reads from the socket and handling errors from the read itself.
 
-    def _read_to_pos(self, pos: int) -> None:
+        This function is greedy and will read as much data as possible
+        from the underlying socket.
+        """
+        # This will raise BlockingIOError if there is no data to read
+        # which we store in the MARSHALL_STREAM_END_ERROR object
+        while True:
+            try:
+                data = self._sock_reader(DEFAULT_BUFFER_SIZE)  # type: ignore[union-attr]
+            except OSError as e:
+                errno = e.errno
+                if errno == EAGAIN or errno == EWOULDBLOCK:
+                    raise MARSHALL_STREAM_END_ERROR
+                raise
+            if data == b"":
+                raise EOFError()
+            self._buf += data
+            if len(self._buf) >= pos:
+                return
+
+    def _read_stream(self, pos: _int, missing_bytes: _int) -> bytes:
+        """Read from the stream."""
+        data = self._stream_reader(missing_bytes)  # type: ignore[misc]
+        if data is None:
+            raise MARSHALL_STREAM_END_ERROR
+        if data == b"":
+            raise EOFError()
+        self._buf += data
+        if len(self._buf) < pos:
+            raise MARSHALL_STREAM_END_ERROR
+
+    def _read_to_pos(self, pos: _int) -> None:
         """
         Read from underlying socket into buffer.
 
@@ -273,19 +357,15 @@ class Unmarshaller:
         :returns:
             None
         """
-        start_len = len(self._buf)
-        missing_bytes = pos - (start_len - self._pos)
+        missing_bytes = pos - len(self._buf)
+        if missing_bytes <= 0:
+            return
         if self._sock is None:
-            data = self._stream_reader(missing_bytes)  # type: ignore[misc]
+            self._read_stream(pos, missing_bytes)
+        elif self._negotiate_unix_fd:
+            self._read_sock_with_fds(pos, missing_bytes)
         else:
-            data = self._read_sock(missing_bytes)
-        if data == b"":
-            raise EOFError()
-        if data is None:
-            raise MARSHALL_STREAM_END_ERROR
-        self._buf += data
-        if len(data) + start_len != pos:
-            raise MARSHALL_STREAM_END_ERROR
+            self._read_sock_without_fds(pos)
 
     def read_uint32_unpack(self, type_: _SignatureType) -> int:
         return self._read_uint32_unpack()
@@ -362,12 +442,12 @@ class Unmarshaller:
             return Variant(SIGNATURE_TREE_N, self._read_int16_unpack(), False)
         elif signature == "ay":
             return Variant(
-                SIGNATURE_TREE_AY, self._read_array(SIGNATURE_TREE_AY_TYPES_0), False
+                SIGNATURE_TREE_AY, self.read_array(SIGNATURE_TREE_AY_TYPES_0), False
             )
         elif signature == "a{qv}":
             return Variant(
                 SIGNATURE_TREE_A_QV,
-                self._read_array(SIGNATURE_TREE_A_QV_TYPES_0),
+                self.read_array(SIGNATURE_TREE_A_QV_TYPES_0),
                 False,
             )
         elif signature == "s":
@@ -378,17 +458,17 @@ class Unmarshaller:
             return Variant(SIGNATURE_TREE_O, self._read_string_unpack(), False)
         elif signature == "as":
             return Variant(
-                SIGNATURE_TREE_AS, self._read_array(SIGNATURE_TREE_AS_TYPES_0), False
+                SIGNATURE_TREE_AS, self.read_array(SIGNATURE_TREE_AS_TYPES_0), False
             )
         elif signature == "a{sv}":
             return Variant(
                 SIGNATURE_TREE_A_SV,
-                self._read_array(SIGNATURE_TREE_A_SV_TYPES_0),
+                self.read_array(SIGNATURE_TREE_A_SV_TYPES_0),
                 False,
             )
         elif signature == "ao":
             return Variant(
-                SIGNATURE_TREE_AO, self._read_array(SIGNATURE_TREE_AO_TYPES_0), False
+                SIGNATURE_TREE_AO, self.read_array(SIGNATURE_TREE_AO_TYPES_0), False
             )
         elif signature == "u":
             return Variant(SIGNATURE_TREE_U, self._read_uint32_unpack(), False)
@@ -417,9 +497,6 @@ class Unmarshaller:
         ), self._readers[type_.children[1].token](self, type_.children[1])
 
     def read_array(self, type_: _SignatureType) -> Iterable[Any]:
-        return self._read_array(type_)
-
-    def _read_array(self, type_: _SignatureType) -> Iterable[Any]:
         self._pos += -self._pos & 3  # align 4 for the array
         self._pos += (
             -self._pos & (UINT32_SIZE - 1)
@@ -469,7 +546,7 @@ class Unmarshaller:
                 while self._pos - beginning_pos < array_length:
                     self._pos += -self._pos & 7  # align 8
                     key = self._read_string_unpack()
-                    result_dict[key] = self._read_array(child_1)
+                    result_dict[key] = self.read_array(child_1)
             else:
                 reader_1 = self._readers[child_1_token]
                 reader_0 = self._readers[child_0_token]
@@ -494,7 +571,7 @@ class Unmarshaller:
             result_list.append(reader(self, child_type))
         return result_list
 
-    def header_fields(self, header_length: int) -> Dict[str, Any]:
+    def _header_fields(self, header_length: _int) -> Dict[str, Any]:
         """Header fields are always a(yv)."""
         beginning_pos = self._pos
         headers = {}
@@ -580,13 +657,15 @@ class Unmarshaller:
         self._msg_len = (
             self._header_len + (-self._header_len & 7) + self._body_len
         )  # align 8
-        self._readers = self._readers_by_type[endian]
+        if self._endian != endian:
+            self._readers = self._readers_by_type[endian]
+            self._endian = endian
 
     def _read_body(self) -> None:
         """Read the body of the message."""
         self._read_to_pos(HEADER_SIGNATURE_SIZE + self._msg_len)
         self._pos = HEADER_ARRAY_OF_STRUCT_SIGNATURE_POSITION
-        header_fields = self.header_fields(self._header_len)
+        header_fields = self._header_fields(self._header_len)
         self._pos += -self._pos & 7  # align 8
         header_fields.pop("unix_fds", None)  # defined by self._unix_fds
         signature = header_fields.pop("signature", "")
@@ -600,24 +679,24 @@ class Unmarshaller:
             tree = SIGNATURE_TREE_SA_SV_AS
             body = [
                 self._read_string_unpack(),
-                self._read_array(SIGNATURE_TREE_SA_SV_AS_TYPES_1),
-                self._read_array(SIGNATURE_TREE_SA_SV_AS_TYPES_2),
+                self.read_array(SIGNATURE_TREE_SA_SV_AS_TYPES_1),
+                self.read_array(SIGNATURE_TREE_SA_SV_AS_TYPES_2),
             ]
         elif signature == "oa{sa{sv}}":
             tree = SIGNATURE_TREE_OA_SA_SV
             body = [
                 self._read_string_unpack(),
-                self._read_array(SIGNATURE_TREE_OA_SA_SV_TYPES_1),
+                self.read_array(SIGNATURE_TREE_OA_SA_SV_TYPES_1),
             ]
         elif signature == "oas":
             tree = SIGNATURE_TREE_OAS
             body = [
                 self._read_string_unpack(),
-                self._read_array(SIGNATURE_TREE_OAS_TYPES_1),
+                self.read_array(SIGNATURE_TREE_OAS_TYPES_1),
             ]
         elif signature == "a{oa{sa{sv}}}":
             tree = SIGNATURE_TREE_A_OA_SA_SV
-            body = [self._read_array(SIGNATURE_TREE_A_OA_SA_SV_TYPES_0)]
+            body = [self.read_array(SIGNATURE_TREE_A_OA_SA_SV_TYPES_0)]
         elif signature == "o":
             tree = SIGNATURE_TREE_O
             body = [self._read_string_unpack()]
@@ -625,9 +704,12 @@ class Unmarshaller:
             tree = get_signature_tree(signature)
             body = [self._readers[t.token](self, t) for t in tree.types]
 
+        flags = MESSAGE_FLAG_MAP.get(self._flag)
+        if flags is None:
+            flags = MESSAGE_FLAG_INTENUM(self._flag)
         self._message = Message(
             message_type=MESSAGE_TYPE_MAP[self._message_type],
-            flags=MESSAGE_FLAG_MAP[self._flag],
+            flags=flags,
             unix_fds=self._unix_fds,
             signature=tree,
             body=body,
@@ -637,6 +719,7 @@ class Unmarshaller:
             validate=False,
             **header_fields,
         )
+        self._read_complete = True
 
     def unmarshall(self) -> Optional[Message]:
         """Unmarshall the message.
@@ -654,6 +737,8 @@ class Unmarshaller:
         if there are not enough bytes in the buffer. This allows unmarshall
         to be resumed when more data comes in over the wire.
         """
+        if self._read_complete:
+            self._next_message()
         try:
             if not self._msg_len:
                 self._read_header()
