@@ -36,7 +36,7 @@ if TYPE_CHECKING:
 
 
 class _Method:
-    def __init__(self, fn, name, disabled=False):
+    def __init__(self, fn, name: str, disabled=False):
         in_signature = ""
         out_signature = ""
 
@@ -71,7 +71,7 @@ class _Method:
         self.out_signature_tree = get_signature_tree(out_signature)
 
 
-def method(name: str = None, disabled: bool = False):
+def method(name: Optional[str] = None, disabled: bool = False):
     """A decorator to mark a class method of a :class:`ServiceInterface` to be a DBus service method.
 
     The parameters and return value must each be annotated with a signature
@@ -149,7 +149,7 @@ class _Signal:
         self.introspection = intr.Signal(self.name, args)
 
 
-def signal(name: str = None, disabled: bool = False):
+def signal(name: Optional[str] = None, disabled: bool = False):
     """A decorator to mark a class method of a :class:`ServiceInterface` to be a DBus signal.
 
     The signal is broadcast on the bus when the decorated class method is
@@ -271,7 +271,7 @@ class _Property(property):
 
 def dbus_property(
     access: PropertyAccess = PropertyAccess.READWRITE,
-    name: str = None,
+    name: Optional[str] = None,
     disabled: bool = False,
 ):
     """A decorator to mark a class method of a :class:`ServiceInterface` to be a DBus property.
@@ -350,9 +350,9 @@ class ServiceInterface:
         self.__properties: List[_Property] = []
         self.__signals: List[_Signal] = []
         self.__buses = set()
-        self.__handlers: dict[
+        self.__handlers: Dict[
             BaseMessageBus,
-            dict[_Method, Callable[[Message, Callable[[Message], None]], None]],
+            Dict[_Method, Callable[[Message, Callable[[Message], None]], None]],
         ] = {}
 
         for name, member in inspect.getmembers(type(self)):
@@ -490,9 +490,7 @@ class ServiceInterface:
 
     @staticmethod
     def _msg_body_to_args(msg: Message) -> List[Any]:
-        if not msg.unix_fds or not signature_contains_type(
-            msg.signature_tree, msg.body, "h"
-        ):
+        if not signature_contains_type(msg.signature_tree, msg.body, "h"):
             return msg.body
 
         # XXX: This deep copy could be expensive if messages are very
@@ -504,7 +502,9 @@ class ServiceInterface:
 
     @staticmethod
     def _fn_result_to_body(
-        result: List[Any], signature_tree: SignatureTree, replace_fds: bool = True
+        result: Optional[Any],
+        signature_tree: SignatureTree,
+        replace_fds: bool = True,
     ) -> Tuple[List[Any], List[int]]:
         """The high level interfaces may return single values which may be
         wrapped in a list to be a message body. Also they may return fds
@@ -532,7 +532,7 @@ class ServiceInterface:
 
     @staticmethod
     def _handle_signal(
-        interface: "ServiceInterface", signal: _Signal, result: List[Any]
+        interface: "ServiceInterface", signal: _Signal, result: Optional[Any]
     ) -> None:
         body, fds = ServiceInterface._fn_result_to_body(result, signal.signature_tree)
         for bus in ServiceInterface._get_buses(interface):
